
-- Lists ------------------------------------------------------------------

List : Set -> Set
nil : (A:Set) -> List A
cons : (A:Set) -> A -> List A -> List A

listInd :  (P	    :	(A:Set) -> List A -> Prop) ->
	    (P_nil  :	(A:Set) -> P A (nil A)) ->
	    (P_cons :	(A:Set) -> (x:A) -> (xs:List A) ->
			P A xs -> P A (cons A x xs)) ->
	    (A	    :	Set) ->
	    (xs	    :	List A) ->
	    P A xs

listInd' : (A	    :	Set) ->
	    (P	    :	List A -> Prop) ->
	    (P_nil  :	P (nil A)) ->
	    (P_cons :	(x:A) -> (xs:List A) ->
			P xs -> P (cons A x xs)) ->
	    (xs	    :	List A) ->
	    P xs

-- The merge package ------------------------------------------------------

A : Set
As : Set
As = List A

nl : List A
nl = nil A

infixr 7 ::
(::) : A -> List A -> List A
(::) = cons A

-- Axioms for <= ----------------------------------------------------------

infix 6 <=, >=, <, >
(<=) : A -> A -> Prop

(>=) : A -> A -> Prop
x >= y = y <= x

(<) : A -> A -> Prop
x < y = Not (y <= x)

(>) : A -> A -> Prop
(>) = \x y -> y < x

leq_trans   : (x,y,z:A) -> Holds (x <= y /\ y <= z ==> x <= z)
leq_total   : (x,y:A) -> Holds (x < y ==> x <= y)

-- zero -------------------------------------------------------------------

zero	    : A
leq_zero    : (x:A) -> Holds (zero <= x)

-- Merge ------------------------------------------------------------------

merge : (xs,ys:As) -> As

merge_nil1	: (ys:As) -> Holds (Eq As (merge nl ys) ys)
merge_nil2	: (xs:As) -> Holds (Eq As (merge xs nl) xs)
merge_cons_cons : (x:A) -> (xs:As) -> (y:A) -> (ys:As) ->
		    Holds (
			( x <= y    ==> Eq As   (merge (x::xs) (y::ys))
						(x :: merge xs (y::ys))
			) /\
			( x > y	    ==> Eq As   (merge (x::xs) (y::ys))
						(y :: merge (x::xs) ys)
			)
		    )

-- << ---------------------------------------------------------------------

infix 6 <<
(<<) : (x:A) -> (xs:As) -> Prop

less_nil    : (x:A) -> Holds (x << nl)
less_cons   : (x:A) -> (y:A) -> (ys:As) ->
		Holds (x <= y /\ x << ys <=> x << y::ys)

-- ordered ----------------------------------------------------------------

ordered : (xs:As) -> Prop

ordered_nil	: Holds (ordered nl)
ordered_cons	: (x:A) -> (xs:As) ->
		    Holds (x << xs /\ ordered xs <=> ordered (x::xs))

-- Lemma 1:: Transitiviy for << --------------------------------------------

Lem1 : (x,y:A) -> (xs:As) -> Prop
Lem1 = \x y xs -> Holds (x <= y /\ y << xs ==> x << xs)

lem1 : (x,y:A) -> (xs:As) -> Lem1 x y xs
lem1 = \x y ->
	let base : Lem1 x y nl
	    base = fol-plugin (less_nil,ordered_nil)

	    step : (z:A) -> (zs:As) -> Lem1 x y zs -> Lem1 x y (z::zs)
	    step = \z zs ih -> fol-plugin	( ih
					, less_cons
					, leq_trans
					)
	in listInd' A (Lem1 x y) base step

-- Lemma 2:: Merge preserves order -----------------------------------------

Lem2 : (xs,ys:As) -> (z:A) -> Prop
Lem2 = \xs ys z ->   Holds ( ordered (z::xs) /\
			    ordered (z::ys) ==>
			    ordered (z::merge xs ys)
			  )
lem2 : (xs,ys:As) -> (z:A) -> Lem2 xs ys z
lem2 =
    let base_xs : (ys:As) -> (z:A) -> Lem2 nl ys z
	base_xs = fol-plugin (merge_nil1, ordered_cons)
	step_xs : (x:A) -> (xs:As) ->
		    ((ys:As) -> (z:A) -> Lem2 xs ys z) ->
		    (ys:As) -> (z:A) -> Lem2 (x::xs) ys z
	step_xs = \x xs ih1 ->
		let base_ys : (z:A) -> Lem2 (x::xs) nl z
		    base_ys = fol-plugin (merge_nil2, ordered_cons)
		    -- Doesn't go through after the change
		    -- of constants
		    step_ys : (y:A) -> (ys:As) ->
				((z:A) -> Lem2 (x::xs) ys z) ->
				(z:A) -> Lem2 (x::xs) (y::ys) z
		    step_ys = \w ws ih2 -> 
			    fol-plugin  ( ih1, ih2, lem1
				    , leq_total
				    , merge_cons_cons
				    , ordered_cons
				    , less_cons
				    )
		in listInd' A (\ys -> (z:A) -> Lem2 (x::xs) ys z) base_ys step_ys
    in listInd' A   (\xs -> (ys:As) -> (z:A) -> Lem2 xs ys z)
		    base_xs step_xs

-- Lemma 3:: zero is very small --------------------------------------------

lem3 : (xs:As) -> Holds (zero << xs)
lem3 =	let base : Holds (zero << nl)
	    base = fol-plugin (less_nil)
	    step : (x:A) -> (xs:As) -> Holds (zero << xs) -> Holds (zero << (x::xs))
	    step = \x xs ih -> fol-plugin (ih, less_cons, leq_zero)
	in listInd' A (\xs -> Holds (zero << xs)) base step

-- Theorem:: merge preserves order -----------------------------------------

thm : (xs,ys:As) -> Holds (ordered xs /\ ordered ys ==> ordered (merge xs ys))
thm = fol-plugin (lem2, lem3, ordered_cons)

