
-- # import Property
-- # import QuickCheck
-- # import List
-- # import Bool.Proofs
-- # import List.Proofs

-- Helpers ----------------------------------------------------------------

null : (A:Set) -> List A -> Bool
null A []     = true
null A (x::xs) = false

head : (A:Set) -> List A -> A
head A (x::xs) = x
head A []     = epsilon A

tail : (A:Set) -> List A -> List A
tail A (x::xs) = xs
tail A []     = epsilon (List A)

orT : (x,y:Bool) -> Holds (x || y) ==> Holds x \/ Holds y
orT true y = fol-plugin()
orT false y = fol-plugin()

nullTrue : (A:Set) -> (xs:List A) -> Holds (null A xs) ==> Eq (List A) xs []
nullTrue A [] = fol-plugin()
nullTrue A (x::xs) = fol-plugin(List)

nullRev : (A:Set) -> (xs:List A) ->
	      Holds (null A xs) ==>
	      Holds (null A (reverse A xs))
nullRev A = fol-plugin(nullTrue A, Holds, rev, null)

-- Queue specification ----------------------------------------------------

Queue = List

empty : (A:Set) -> Queue A
empty A = []

isEmpty : (A:Set) -> Queue A -> Bool
isEmpty = null

add : (A:Set) -> A -> Queue A -> Queue A
add A x q = append A q [x]

remove : (A:Set) -> Queue A -> Queue A
remove = tail

front : (A:Set) -> Queue A -> A
front = head

-- Queue implementation ---------------------------------------------------

QueueI : (A:Set) -> Set
QueueI A = {List A; List A}

flipQ : (A:Set) -> QueueI A -> QueueI A
flipQ A (f,b) = if (QueueI A) (null A f)
		    (reverse A b, [])
		    (f, b)

emptyI : (A:Set) -> QueueI A
emptyI A = ([], [])

isEmptyI : (A:Set) -> QueueI A -> Bool
isEmptyI A (f,_) = null A f

addI : (A:Set) -> A -> QueueI A -> QueueI A
addI A x (f,b) = flipQ A (f, x::b)

frontI : (A:Set) -> QueueI A -> A
frontI A (x::f,b)  = x
frontI A ([],b)	  = epsilon A

removeI : (A:Set) -> QueueI A -> QueueI A
removeI A (x::f,b) = flipQ A (f,b)
removeI A ([], b) = epsilon (QueueI A)

-- Retrieving the specification -------------------------------------------

retrieve : (A:Set) -> QueueI A -> Queue A
retrieve A (f,b) = append A f (reverse A b)

-- Generators -------------------------------------------------------------

arbQueue : Gen (Queue Nat)
arbQueue = arbList Nat arbNat

arbQueueI : Gen (QueueI Nat)
arbQueueI = arbSigma (List Nat) (\_ -> List Nat)
		(arbList Nat arbNat)
		(\_ -> arbList Nat arbNat)

-- Properties -------------------------------------------------------------

invariant : (A:Set) -> QueueI A -> Prop
invariant A (f,b) = Holds (null A b || not (null A f))

-- retrieve emptyI == empty
PropEmpty : (A:Set) -> Prop
PropEmpty A = Eq (Queue A) (retrieve A (emptyI A)) (empty A)

--testEmpty : (A:Set) -> PropEmpty A
--testEmpty A = qc-plugin()

propEmpty : (A:Set) -> PropEmpty A
propEmpty A = reflexivity (Queue A) []

-- invariant emptyI
PropInvEmpty : (A:Set) -> Prop
PropInvEmpty A = invariant A (emptyI A)

--testInvEmpty : (A:Set) -> PropInvEmpty A
--testInvEmpty A = qc-plugin()

propInvEmpty : (A:Set) -> PropInvEmpty A
propInvEmpty A = tt

-- isEmptyI == isEmpty
PropIsEmpty : (A:Set) -> Prop
PropIsEmpty A = (q:QueueI A) ->
		  invariant A q ==>
		  Eq Bool (isEmptyI A q)
			  (isEmpty A (retrieve A q))

--testIsEmpty : PropIsEmpty Nat
--testIsEmpty = qc-plugin(arbQueueI)

propIsEmpty : (A:Set) -> PropIsEmpty A
propIsEmpty A ([],b)  = fol-plugin(Holds, Bool, propHoldsOr, rev, nullTrue A, propHoldsTrue)
propIsEmpty A (x::f,b) = \_ -> reflexivity Bool false

-- addI == add
PropAdd : (A:Set) -> Prop
PropAdd A = (x:A) -> (q:QueueI A) ->
	      Eq (Queue A)  (retrieve A (addI A x q))
			    (add A x (retrieve A q))

--testAdd : PropAdd Nat
--testAdd = qc-plugin(arbNat, arbQueueI)

propAdd : (A:Set) -> PropAdd A
propAdd A x (y::f,b) = fol-plugin(propRevApp A, propAppAssoc A, rev, append)
propAdd A x ([],b)  = fol-plugin(propRevApp A, rev, append, propAppNil A)

-- invariant addI
--   (add actually establishes the invariant even if its argument
--    doesn't satisfies it)
PropInvAdd : (A:Set) -> Prop
PropInvAdd A = (x:A) -> (q:QueueI A) ->
		  invariant A (addI A x q)

--testInvAdd : PropInvAdd Nat
--testInvAdd = qc-plugin(arbNat, arbQueueI)

propInvAdd : (A:Set) -> PropInvAdd A
propInvAdd A x (y::f,b)	= tt
propInvAdd A x ([], b)	= tt

-- frontI == front
PropFront : (A:Set) -> Prop
PropFront A = (q:QueueI A) ->
		Holds (not (isEmptyI A q)) ==>
		  Eq A (frontI A q) (front A (retrieve A q))

--testFront : PropFront Nat
--testFront = qc-plugin(arbQueueI)

propFront : (A:Set) -> PropFront A
propFront A (x::f,b) = \_ -> reflexivity A x
propFront A ([], b) = fol-plugin()

-- frontI == front (including bottom cases)
PropFront' : (A:Set) -> Prop
PropFront' A = (q:QueueI A) ->
		invariant A q ==>
		  Eq A (frontI A q) (front A (retrieve A q))

--testFront' : PropFront' Nat
--testFront' = qc-plugin(arbQueueI)

propFront' : (A:Set) -> PropFront' A
propFront' A (x::f,b) = \_ -> reflexivity A x
propFront' A ([], b) = fol-plugin(Holds, orT, rev, nullTrue A, head)

-- removeI == remove
PropRemove : (A:Set) -> Prop
PropRemove A = (q:QueueI A) ->
		Holds (not (isEmptyI A q)) ==>
		  Eq (Queue A)
		    (retrieve A (removeI A q))
		    (remove A (retrieve A q))

--testRemove : PropRemove Nat
--testRemove = qc-plugin(arbQueueI)

propRemove : (A:Set) -> PropRemove A
propRemove A ([], b)	= fol-plugin()
propRemove A ([x],b)	= fol-plugin(propAppNil A)
propRemove A (x::y::f, b) = fol-plugin()

-- invariant removeI
--   (remove actually establishes the invariant even if its argument
--    doesn't satisfies it)
PropRemoveInv : (A:Set) -> Prop
PropRemoveInv A = (q:QueueI A) ->
		    Holds (not (isEmptyI A q))	==>
		    invariant A (removeI A q)

--testRemoveInv : PropRemoveInv Nat
--testRemoveInv = qc-plugin(arbQueueI)

propRemoveInv : (A:Set) -> PropRemove A
propRemoveInv A ([], b)	  = fol-plugin()
propRemoveInv A ([x],b)	  = fol-plugin(propAppNil A)
propRemoveInv A (x::y::f,b) = fol-plugin()

