
-- # import Property
-- # import CoverPrelude
-- # import Tree
-- # import GHC.Base


{------------------------------------------------------------------------

  First we define the set of total and finite trees, together
  with an injection function into the real Haskell trees.

 ------------------------------------------------------------------------}

data Tree (a : Set) : Set where
  Leaf	: a -> Tree a
  Node	: Tree a -> Tree a -> Tree a
  Bot	: Tree a

injTree (a : ISet) : Tree a -> "Tree::Tree" a
injTree (Leaf x)    = "Tree.Leaf" x
injTree (Node l r)  = "Tree.Node" (injTree l) (injTree r)
injTree Bot	    = "_|_"


{------------------------------------------------------------------------

  Models of the Haskell functions

 ------------------------------------------------------------------------}

id (a : ISet) : a -> a
id x = x

model_id (a : ISet) : id {a} == "GHC.Base.id" {a}
model_id = impE (extensionality id "GHC.Base.id") cover-plugin(include "GHC.Base")

map (a, b : Set) : (a -> b) -> Tree a -> Tree b
map f (Leaf x)	  = Leaf (f x)
map f (Node l r)  = Node (map f l) (map f r)
map f Bot	  = Bot

model_map (a, b : ISet) : (f : a -> b) -> (t : Tree a) ->
    injTree (map f t) == "Tree.mapTree" f (injTree t)
model_map f (Leaf x)	= cover-plugin(include "Tree")
model_map f (Node l r)	= cover-plugin(include "Tree", model_map f l, model_map f r)
model_map f Bot		= cover-plugin(include "Tree")

(@) (a, b, c : ISet) : (b -> c) -> (a -> b) -> a -> c
(@) f g x = f (g x)

model_compose (a, b, c : ISet) : (f : b -> c) -> (g : a -> b) ->
				   (x : a) -> (f @ g) x == "GHC.Base.." f g x
model_compose f g = cover-plugin(include "GHC.Base")


{------------------------------------------------------------------------

  Tree/mapTree is a functor.

 ------------------------------------------------------------------------}

prop_mapId (a : Set) : (t : Tree a) -> map id t == t
prop_mapId (Leaf x)   = cover-plugin()
prop_mapId (Node l r) = cover-plugin(prop_mapId l, prop_mapId r)
prop_mapId Bot	      = cover-plugin()

prop_mapCompose (a, b, c : Set) :
    (f : b -> c) -> (g : a -> b) -> (t : Tree a) ->
    map (f @ g) t == map f (map g t)
prop_mapCompose f g (Leaf x)	= fol-plugin()
prop_mapCompose f g (Node l r)	= fol-plugin( prop_mapCompose f g l
					    , prop_mapCompose f g r
					    )
prop_mapCompose f g Bot		= fol-plugin()



-- prop_mapId' (a : ISet) :
--     (t : Tree a) -> "Tree.mapTree" "GHC.Base.id" (injTree t) == injTree t
-- prop_mapId' (Leaf x) = ? {-fol-plugin( prop_mapId t
-- 			  , model_map
-- 			  , model_id
-- 			  , freeze id
-- 			  )-}

-- prop_mapId (Node l r)	= cover-plugin( include "Tree", prop_mapId l
-- 				      , prop_mapId r
-- 				      )
-- prop_mapId Bot		= cover-plugin(include "Tree")

{-
prop_mapCompose (a, b, c : ISet) :
    (f : b -> c) -> (g : a -> b) -> (t : Tree a) ->
    map (f @ g) (injTree t) == map f (map g (injTree t))
prop_mapCompose f g (Leaf x)   = cover-plugin( include "Tree", include "GHC.Base" )
prop_mapCompose f g (Node l r) = cover-plugin( include "Tree", include "GHC.Base"
					     , prop_mapCompose f g l
					     , prop_mapCompose f g r
					     )
-}
