
-- # import Set

-- Fixities ---------------------------------------------------------------

infix  5 <==>
infixr 6 ==>
infixr 7 /\, \/

-- PROP -------------------------------------------------------------------

sig PROP : Type where
  prf : Prop

Prf : PROP -> Prop
Prf P = P.prf

mkPROP : Prop -> PROP
mkPROP P = struct prf = P

infix 0 `by`
by : (P:PROP) -> Prf P -> Prf P
by _ p = p

-- True and False ---------------------------------------------------------

data True'  : Prop where tt : True'
data False' : Prop where

True  = mkPROP True'
False = mkPROP False'

falseE (P:PROP) : Prf False -> Prf P

-- Implies ----------------------------------------------------------------

(==>) : PROP -> PROP -> PROP
P ==> Q = mkPROP (Prf P -> Prf Q)

impI (P,Q:PROP) : (Prf P -> Prf Q) -> Prf (P ==> Q)
impI f = f

impE (P,Q:PROP) : Prf (P ==> Q) -> (Prf P -> Prf Q)
impE f = f

-- Not --------------------------------------------------------------------

Not : PROP -> PROP
Not P = P ==> False

notI (P:PROP) : (Prf P -> Prf False) -> Prf (Not P)
notI = impI {_} {False}
  -- Annoying. Solution: don't treat A -> B as (x:A) -> B   [x fresh]

notE (P:PROP) : Prf (Not P) -> Prf P -> Prf False
notE = impE {_} {False}

-- Or ---------------------------------------------------------------------

data (\/') (P,Q:Prop) : Prop where
  orI1 : P -> P \/' Q
  orI2 : Q -> P \/' Q

(\/) : PROP -> PROP -> PROP
P \/ Q = mkPROP (Prf P \/' Prf Q)

orE (P, Q, R : PROP) : Prf (P \/ Q) -> (Prf P -> Prf R) -> (Prf Q -> Prf R) -> Prf R
orE (orI1 p) l r = l p
orE (orI2 q) l r = r q

-- And --------------------------------------------------------------------

data (/\') (P,Q:Prop) : Prop where
  andI : P -> Q -> P /\' Q

(/\) : PROP -> PROP -> PROP
P /\ Q = mkPROP (Prf P /\' Prf Q)

andE1 (P, Q : PROP) : Prf (P /\ Q) -> Prf P
andE1 (andI p q) = p

andE2 (P, Q : PROP) : Prf (P /\ Q) -> Prf Q
andE2 (andI p q) = q

-- Equiv ------------------------------------------------------------------

(<==>) : PROP -> PROP -> PROP
P <==> Q = (P ==> Q) /\ (Q ==> P)

equivIrefl (P : PROP) : Prf (P <==> P)
equivIrefl = andI (\p -> p) (\p -> p)

-- Forall -----------------------------------------------------------------

ForAll (A : SET) : (El A -> PROP) -> PROP
ForAll {A} B = mkPROP ((x : El A) -> Prf (B x))

forAllI (A : SET)(B : El A -> PROP) :
  ((x:El A) -> Prf (B x)) -> Prf (ForAll B)
forAllI f = f

forAllE (A : SET)(B : El A -> PROP) :
  Prf (ForAll B) -> (x : El A) -> Prf (B x)
forAllE f = f
