
-- # import Set
-- # import Prop
-- # import Pred

-- Relation ---------------------------------------------------------------

sig Rel (A:SET) : Type where
  rel : El A -> El A -> PROP

mkRel (A:SET) : (El A -> El A -> PROP) -> Rel A
mkRel rel = struct rel = rel

-- Special kinds of relations ---------------------------------------------

Reflexive (A:SET) : Rel A -> PROP
Reflexive R = ForAll $ \x -> R.rel x x

sig ReflexiveRel (A:SET) : Type extends Rel where
  refl : Prf (Reflexive (mkRel rel))

Symmetric (A:SET) : Rel A -> PROP
Symmetric R = ForAll $ \x -> ForAll $ \y ->
	      R.rel x y ==> R.rel y x

sig SymmetricRel (A:SET) : Type extends Rel where
  sym : Prf (Symmetric (mkRel rel))

symmetry (A:SET)(R:SymmetricRel A)(x,y:El A) :
  Prf (R.rel x y ==> R.rel y x)
symmetry {_} {R} xy = R.sym _ _ xy

Asymmetric (A:SET) : Rel A -> PROP
Asymmetric R =
  ForAll $ \x ->
  ForAll $ \y ->
    R.rel x y ==> Not (R.rel y x)

sig AsymmetricRel (A:SET) : Type extends Rel where
  asym : Prf (Asymmetric (mkRel rel))

Transitive (A:SET) : Rel A -> PROP
Transitive R =
  ForAll $ \x ->
  ForAll $ \y ->
  ForAll $ \z ->
    R.rel x y ==> R.rel y z ==> R.rel x z

sig TransitiveRel (A:SET) : Type extends Rel where
  trans : Prf (Transitive (mkRel rel))

transitivity (A:SET)(R:TransitiveRel A)(x,y,z:El A) :
  Prf (R.rel x y ==> R.rel y z ==> R.rel x z)
transitivity {_} {R} xy yz = R.trans _ _ _ xy yz

sig EquivalenceRel (A:SET) : Type
  extends ReflexiveRel
	, SymmetricRel
	, TransitiveRel
  where

Total (A:SET) : Rel A -> PROP
Total R = ForAll $ \x -> ForAll $ \y ->
	  R.rel x y \/ R.rel y x

-- A total relation is also reflexive so it makes sense to inherit from
-- ReflexiveRel.
sig TotalRel (A:SET) : Type extends ReflexiveRel where
  trichotomy : Prf (Total (mkRel rel))

Substitutive (A:SET) : Rel A -> Pred A -> PROP
Substitutive R P =
  ForAll $ \x ->
  ForAll $ \y ->
    R.rel x y ==> P.pred x ==> P.pred y

sig SubstitutiveRel (A:SET) : Type extends Rel where
  subst : (P : Pred A) -> Prf (Substitutive (mkRel rel) P)

-- The empty relation -----------------------------------------------------

sig SymTransSubstRel (A:SET) : Type
  extends SymmetricRel
	, TransitiveRel
	, SubstitutiveRel where

emptyRel (A:SET) : SymTransSubstRel A
emptyRel =
  struct rel	 x y	     = False
	 sym	 x y   xy    = xy
	 trans	 x y z xy yz = xy
	 subst P x y   xy px = falseE xy

-- Some nice properties ---------------------------------------------------

(===) (A:SET) : Rel A -> Rel A -> PROP
R === S = ForAll $ \x -> ForAll $ \y ->
	  R.rel x y <==> S.rel x y

(=<=) (A:SET) : Rel A -> Rel A -> PROP
R =<= S = ForAll $ \x -> ForAll $ \y ->
	  R.rel x y ==> S.rel x y

{-
totalRefl (A:SET) : (R : Rel A) -> Prf (Total R ==> Reflexive R)
totalRefl R trich x = orE (trich _ _) (\p -> p) (\p -> p)
-- Doesn't work! Problem: meta variables not allowed in with expression.
-- totalRefl {A} R trich x with trich _ _ =>  
-- 	  {A} R trich x (orI1' p) = p
-- 	  {A} R trich x (orI2' p) = p

TotalReflRel (A:SET) : (R : TotalRel A) -> { R' : ReflexiveRel A; Prf (R' === R) }
TotalReflRel R =
  ( struct rel  = R.rel
	   refl = totalRefl R R.trichotomy
  , \_ _ -> equivIrefl
  )
-}
