{-# OPTIONS -cpp #-}

{-| Import chasing. Requires the variable @_ROOT_@ to be cpp'd to the root
    directoty of the agdaLight distribution.
-}
module Interface.ChaseImports where

import Control.Monad
import Data.Char
import Data.List
import System.Directory

import Utils.Monads

{-|
    @chaseImports include_dirs files@.
    Chase imports by looking for lines of the form @-- # import foo.bar.baz@
    The interpretation is that there should be a file foo\/bar\/baz.agda in
    either the same directory as the file containing the import, in the current
    working directory, or in the specified include directories (agdaLight\/lib
    is default).
-}
chaseImports :: [FilePath] -> [FilePath] -> IO [FilePath]
chaseImports is files =
    do	cwd <- addSlash <$> getCurrentDirectory
	(nub . concat)
	    <$> mapM (\file -> findImports cwd includes [file]) files
    where
	includes = map addSlash is

canonify (x:xs) (y:ys)
    | x == y	    = canonify xs ys
canonify [] ys	    = ys
canonify ('/':_) ys = ys
canonify xs ys	    = dotdot xs ++ ys

dotdot []	= []
dotdot ('/':xs) = '/' : dotdot xs
dotdot xs	=
    case break (== '/') xs of
	(_, xs)	-> ".." ++ dotdot xs

findImports :: FilePath -> [FilePath] -> [FilePath] -> IO [FilePath]
findImports cwd includes prev_files@(file:_)
    | isCycle prev_files    = fail $ "While chasing imports: Cycle detected:\n"
				    ++ unlines (map ("  " ++ ) prev_files)
    | otherwise		    =
	do  imps	<- (concatMap findImport . lines) <$> readFile file
	    let paths = getPath file : "" : includes
	    files	<- map (canonify cwd) <$> mapM (getFileName paths) imps
	    all_files	<- concat <$>
			    mapM (\file -> findImports cwd includes (file:prev_files))
				files
	    return $ all_files ++ [file]
    where
	findImport s =
	    do	("--", s0)	<- lex s
		("#", s1)	<- lex s0
		("import", s2)	<- lex s1
		return $ takeWhile (not . isSpace) $ dropWhile isSpace s2

addSlash "" = ""
addSlash "/" = "/"
addSlash [c] = c:"/"
addSlash (c:s) = c : addSlash s

isCycle xs = nub xs /= xs

getPath :: FilePath -> FilePath
getPath file = reverse $ dropWhile (/='/') $ reverse file

getFileName :: [FilePath] -> String -> IO FilePath
getFileName paths imp =
    foldr (\file getFN ->
	    ifM (doesFileExist file)
		(return file)
		getFN
	)   (fail $ "While chasing imports: File not found " ++ file)
	    (concatMap (\path -> map (path++) [file,lfile]) paths)
    where
	file	= map unDot imp ++ ".agda"
	lfile	= map unDot imp ++ ".lagda"

	unDot '.'   = '/'
	unDot c	    = c

