{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.UnfoldConstants where

import Control.Monad
import Data.Monoid

import Syntax.Internal
import TypeChecker.Values
import Utils.Monads
import Utils.Pairs

-- Unfolding constants ----------------------------------------------------

data UnfoldStrategy = TopLevel | Recursive

-- | Unfolds top level constants. Calls 'genUnfoldConstApp' with 
--   a stop condition that yells as soon as it sees something
--   that isn't a constant.
unfoldConstants :: Value -> Value
unfoldConstants v =
    genUnfoldConstants TopLevel (\_ _ _ -> False) (const Nothing) v

-- | Unfold all constants everywhere.
unfoldAllConstants :: Value -> Value
unfoldAllConstants v =
    genUnfoldConstants Recursive (\_ _ _ -> False) (const Nothing) v

-- | The most general constant unfolding function.
--   Takes a flag indicating if we should only unfold top-level
--   and a don't unfold me condition.
genUnfoldConstants ::
    UnfoldStrategy		->  -- TopLevel or Recursive
    (Name -> DefValue -> [Value] -> Bool)
				->  -- don't unfold me
    (Value -> Maybe Value)	->  -- custom unfold rules
    Value -> Value
genUnfoldConstants strategy nounfold special v
    | Just v' <- special v  = recCall v'
    | otherwise		    =
	case vAppView v of
	    VAVDef x d vs
		| nounfold x d vs   -> VDef x d `vApps` vs'
		| otherwise	    ->
		    case d of
			Definition v -> recCall $ v `vApps` vs
			_	     -> VDef x d `vApps` vs'
		where
		    vs'		= map recCall vs

	    VAVId x t vs   -> VId x t `vApps` map recCall vs
	    VAVLam cl t    -> VLam
				(cl { clApp = \x -> recCall (clApp cl x) } )
				t
	    VAVPi v cl t   -> VPi (recCall v)
				(cl { clApp = \x -> recCall (clApp cl x) } )
				t
	    _		    -> v
    where
	recCall = case strategy of
		    TopLevel	-> id
		    Recursive	-> recCall_

	recCall_ = genUnfoldConstants strategy nounfold special

