{-# OPTIONS -fglasgow-exts -cpp #-}

module Syntax.Sorts where

import Data.Generics
import Data.IORef

import Syntax.Name

import Utils.Pairs
import Utils.Ref

type SortIndexRef = IORef RefValue

data RefValue = Initialized SortIndex
	      | Uninitialized RefType [SortConstraint] [SortIndexRef]
    deriving (Typeable, Data)

data RefType = Global Name | Polymorphic
    deriving (Typeable, Data)

-- | Note: We need greater than to be able to update the constraints when
--   a variable is instantiated.
data SortConstraint = LessEq SortIndex SortIndex
    deriving (Typeable, Data, Eq)


data SortIndex	= SIVar Int Int
		| SIMax [SortIndex]
		| SIMetaVar SortIndexRef Int
    deriving (Typeable, Data, Eq)

siAdd :: SortIndex -> Int -> SortIndex
siAdd (SIVar k n) m	= SIVar k (n + m)
siAdd (SIMetaVar r n) m	= SIMetaVar r (n + m)
siAdd (SIMax is) m	= SIMax $ map (flip siAdd m) is

siSucc = flip siAdd 1

siMax s1 s2 = SIMax $ getMaxs s1 ++ getMaxs s2

siMax' [s]  = s
siMax' ss   = SIMax ss

getMaxs (SIMax ss)  = ss
getMaxs s	    = [s]

siMaxs ss = case flatten ss of
		SIMax ss'   -> siMax' $ simplify ss'
		_	    -> error $ "impossible: siMaxs"
    where
	flatten = foldr siMax (SIMax [])
    
	simplify ss@(SIVar k _ : _) = SIVar k n : simplify ss'
	    where
		(n, ss') = getMax k ss

		getMax k (SIVar k' n : ss)
		    | k == k'	    = (max n -*- id) $ getMax k ss
		getMax k (s : ss)   = (id -*- (s:)) $ getMax k ss
		getMax _ []	    = (minBound, [])

	simplify ss@(SIMetaVar r _ : _) = SIMetaVar r n : simplify ss'
	    where
		(n, ss') = getMax r ss

		getMax r (SIMetaVar r' n : ss)
		    | r == r'	    = (max n -*- id) $ getMax r ss
		getMax r (s : ss)   = (id -*- (s:)) $ getMax r ss
		getMax _ []	    = (minBound, [])

	simplify [] = []

	simplify _ = error $ "impossible: siMaxs.simplify"

