-- Kleene algebra, following Dexter Kozen

D : Set
(+) : D -> D -> D
(<=) : D -> D -> Prop
(&) : D -> D -> D
star : D -> D
Z : D
Eps : D

defLeq : (x:D) -> (y:D) -> (Eq D (x+y) y  <=>   x <= y)


axK : (x:D) -> (y:D) -> (z:D) ->
          (    Eq D (x + y) (y + x)
           /\  Eq D (x+ Z) x
           /\  Eq D (x+ x) x
           /\  Eq D (x + (y+z)) ((x+y)+z)
           /\  Eq D (x & Eps) x
           /\  Eq D (Eps & x) x
           /\  Eq D (x & Z) Z
           /\  Eq D (Z & x) Z
           /\  Eq D (x & (y & z)) ((x & y) & z)
           /\  Eq D (Eps + (x & (star x))) (star x)
           /\  Eq D (Eps + ((star x) & x)) (star x))

rule1 : (x:D) -> (y:D) -> (z:D) -> y + (x & z) <= z ==> ((star x) & y) <= z
rule2 : (x:D) -> (y:D) -> (z:D) -> y + (z & x) <= z ==> (y & (star x)) <= z


transLeq : (x:D) -> (y:D) -> (z:D) -> x <= y /\ y <= z ==> x <= z
transLeq = \ x -> \ y -> \ z -> fol-plugin(defLeq,axK)

ref1Leq : (x:D) -> x <= x
ref1Leq = \ x -> fol-plugin(defLeq,axK)

ref2Leq : (x:D) -> (y:D) -> (x <= y /\ y <= x) -> Eq D x y
ref2Leq = \ x -> \ y -> \ h ->
          let
           rem1 : Eq D (x + y) y
           rem1 = fol-plugin(defLeq,h)
           rem2 : Eq D (y + x) x
           rem2 = fol-plugin(defLeq,h)
          in
           fol-plugin(rem1,rem2,axK)

refLeq : (x:D) -> (y:D) -> ((x <= y /\ y <= x) <=> Eq D x y)
refLeq = \ x -> \ y -> fol-plugin(ref1Leq,ref2Leq)

remPlus : (x:D) -> (y:D) -> (z:D) -> (Eq D (x+y) z  ==>   x<= z /\ y <= z)
remPlus = \ x -> \ y -> \ z -> fol-plugin(defLeq,axK)

lemPlus : (x:D) -> (y:D) -> (z:D) -> (x+y <= z  <=>   x<= z /\ y <= z)
lemPlus = \ x -> \ y -> \ z -> fol-plugin(defLeq,axK,remPlus)

lemConLeqL : (x:D) -> (y:D) -> (z:D) -> (z <= x ==> (z & y) <= (x & y))
lemConLeqL = \ x -> \ y -> \ z -> fol-plugin(defLeq,axK)

lemConLeqR : (x:D) -> (y:D) -> (z:D) -> (z <= x ==> (y & z) <= (y & x))
lemConLeqR = \ x -> \ y -> \ z -> fol-plugin(defLeq,axK)

lemStarLeq : (x:D) -> Eps <= (star x)
lemStarLeq = \ x -> fol-plugin(remPlus,axK)

lem1 : (x:D) -> Eq D ((star x) & (star x)) (star x)
lem1 = \ x ->
       let 
         rem1 : ((star x) & (star x)) <= star x
         rem1 = let
                  rem11 : (x & (star x)) <= star x
                  rem11 = fol-plugin(axK,remPlus)
                  rem12 : star x + (x & (star x)) <= star x
                  rem12 = fol-plugin(axK,defLeq,rem11,lemPlus)
                in fol-plugin(axK,rule1,rem12)
         rem2 : star x <=  ((star x) & (star x))
         rem2 = fol-plugin(axK,lemConLeqL,lemStarLeq)
       in fol-plugin(defLeq,refLeq,rem1,rem2)

remStar : (x:D) -> x <= star x
remStar = \ x ->
           let 
             rem1 : x & (star x) <= star x
             rem1 = fol-plugin(axK,defLeq)
             rem2 : (x & Eps) <= (x & (star x))
             rem2 = fol-plugin(lemStarLeq,lemConLeqR,refLeq)
           in
            fol-plugin(axK,transLeq,rem1,rem2)

lem2 : (x:D) -> Eq D (star (star x)) (star x)
lem2 = \ x ->
        let
          rem1 : Eps & star (star x) <= star x
          rem1 = let
                  rem11 : Eps <= star x 
                  rem11 = lemStarLeq x
                  rem12 : ((star x) & (star x)) <= star x
                  rem12 = fol-plugin(lem1,refLeq)
                  rem13 : Eps + ((star x) & (star x)) <= star x
                  rem13 = fol-plugin(rem11,rem12,lemPlus)
                 in 
                  fol-plugin(axK,rule2,rem13)
          rem2 : (star x) <= star (star x)
          rem2 = fol-plugin(remStar)
        in 
          fol-plugin(rem1,rem2,defLeq,transLeq,axK)

lem3 : (x:D) -> (y:D) -> Eq D ((star (x & y)) & x) (x & (star (y & x)))
lem3 = \ x -> \ y -> 
        let
         rem1 : ((star (x & y)) & x) <= (x & (star (y & x)))
         rem1 = 
          let
            rem11 :  (x + (x & y) & (x & (star (y & x)))) <= (x & (star (y & x))) 
            rem11 = fol-plugin(axK)
          in
           fol-plugin(axK,rule1,rem11)
         rem2 : (x & (star (y & x))) <= ((star (x & y)) & x) 
         rem1 = fol-plugin(axK)
        in
         fol-plugin(rem1,rem2,refLeq)
