-- an example from Michael Beeson

-- # import Property

-- repair Eq::

Eq : (A:Set) -> A -> A -> Prop
Eq A = (==)

-- some logical axioms

Ex : (A:Set) -> (A->Prop) -> Prop
witness : (A:Set) -> (P:A->Prop) -> (x:A) -> P x -> Ex A P
elimEx : (A:Set) -> (P:A->Prop) -> Ex A P -> (C:Prop) -> ((x:A)->P x ->C) -> C

-- natural numbers

data Nat : Set where
 zero : Nat
 succ : Nat -> Nat

indNat : (P:Nat -> Prop) -> P zero -> ((n:Nat) -> P n -> P (succ n)) -> (x:Nat) -> P x
indNat P a g zero = a
indNat P a g (succ n) = g n (indNat P a g n)

-- chain of equalities

data chain (X:Set) (R:X->X->Prop) (x:X) : Prop where
 dirC : (y:X) -> R x y -> chain X R x
 stepC : (y:X) -> R x y -> chain X R y -> chain X R x

last : (X:Set) -> (R:X->X->Prop) -> (x:X) -> chain X R x -> X
last X R x (dirC y p) = y
last X R x (stepC y p q) = last X R y q

trans : (X:Set) -> (X->X->Prop) -> Prop
trans X R = (x:X) -> (y:X) -> (z:X) -> (R x y /\ R y z) ==> R x z

tacChain : (X:Set) -> (x:X) -> (p:chain X (Eq X) x) -> Eq X x (last X (Eq X) x p)
tacChain X x (dirC y p) = p
tacChain X x (stepC y p q) = fol-plugin(p,tacChain X y q)

-- general definitions

isRing : (X:Set) -> (X->X->X) -> (X->X->X) -> (X->X) -> X -> X -> Prop

isRing R (+) (&) minus Zero One =
 (x:R) -> (y:R) -> (z:R) ->
  (   Eq R (x + y) (y + x)
   /\ Eq R (x + Zero) x
   /\ Eq R (x + (minus x)) Zero
   /\ Eq R (x + (y+z)) ((x+y)+z)
   /\ Eq R (x & (y+z)) ((x&y) + (x&z))
   /\ Eq R ((y+z)& x) ((y&x) + (z&x))
   /\ Eq R (x & One) x
   /\ Eq R (One & x) x
   /\ Eq R (x & (y&z)) ((x&y)&z)
  )

-- an example

R : Set
(+) : R -> R -> R
(&) : R -> R -> R
minus : R -> R 
Zero : R
One : R

axR : isRing  R (+) (&) minus Zero One 

-- (==) : R -> R -> Prop
-- (==) = Eq R

infix 14 &
infix 10 +
-- infix 6 ==

lem1 : (x:R) -> ((x & Zero) + (x & One)) == x 
lem1 x =
 tacChain R 
          ((x & Zero) + (x & One))
       (stepC (x & (Zero + One))
                   fol-plugin(axR)
       (stepC (x & One)         
                   fol-plugin(axR)
       (dirC x
                   fol-plugin(axR))))

lem2 : (a:R) -> (b:R) -> ((a+b)&(a+b)) == (((a&a) + (a&b)) + ((b&a) + (b&b)))
lem2 a b =
 tacChain R 
              ((a+b)&(a+b))
       (stepC ((a+b)&a + (a+b)&b)
                   fol-plugin(axR)
       (stepC ((a&a + b&a) + (a&b+b&b))
                   fol-plugin(axR)
       (dirC  ((a&a + a&b) + (b&a+b&b))
                   fol-plugin(axR))))
